'use strict';



var https = require('https');
var util = require('util');
var OriginalAgent = require('_http_agent').Agent;
var OriginalHttpsAgent = https.Agent;
var HttpsAgent;

function Agent(options) {
    if (!(this instanceof Agent)) {
        return new Agent(options);
    }

    options = options || {};
    options.keepAlive = options.keepAlive !== false;
    // default is keep-alive and 15s free socket timeout
    if (options.keepAliveTimeout === undefined) {
        options.keepAliveTimeout = 15000;
    }
    // default timeout is double keepalive timeout
    if (options.timeout === undefined) {
        options.timeout = options.keepAliveTimeout * 2;
    }

    OriginalAgent.call(this, options);

    var self = this;
    self.createSocketCount = 0;
    self.closeSocketCount = 0;
    // socket error event count
    self.errorSocketCount = 0;
    self.requestCount = 0;
    self.timeoutSocketCount = 0;
    self.on('free', function () {
        self.requestCount++;
    });
    self.on('timeout', function () {
        self.timeoutSocketCount++;
    });
    self.on('close', function () {
        self.closeSocketCount++;
    });
    self.on('error', function () {
        self.errorSocketCount++;
    });
}

util.inherits(Agent, OriginalAgent);

Agent.prototype.createSocket = function (req, options) {
    var socket = OriginalAgent.prototype.createSocket.call(this, req, options);
    if (this.keepAlive) {
        // Disable Nagle's algorithm: http://blog.caustik.com/2012/04/08/scaling-node-js-to-100k-concurrent-connections/
        // http://fengmk2.com/benchmark/nagle-algorithm-delayed-ack-mock.html
        socket.setNoDelay(true);
    }
    this.createSocketCount++;
    return socket;
};

Agent.prototype.getCurrentStatus = function () {
    return {
        createSocketCount: this.createSocketCount,
        closeSocketCount: this.closeSocketCount,
        errorSocketCount: this.errorSocketCount,
        timeoutSocketCount: this.timeoutSocketCount,
        requestCount: this.requestCount,
        freeSockets: inspect(this.freeSockets),
        sockets: inspect(this.sockets),
        requests: inspect(this.requests)
    };
};

function inspect(obj) {
    var res = {};
    for (var key in obj) {
        res[key] = obj[key].length;
    }
    return res;
}


var HttpAgent = Agent;
var OriginalHttpsAgent = https.Agent;


HttpsAgent = function HttpsAgent(options) {
    HttpAgent.call(this, options);
    this.defaultPort = 443;
    this.protocol = 'https:';
    this.maxCachedSessions = this.options.maxCachedSessions;
    if (this.maxCachedSessions === undefined)
        this.maxCachedSessions = 100;

    this._sessionCache = {
        map: {},
        list: []
    };
};


util.inherits(HttpsAgent, HttpAgent);

[
    'createConnection',
    'getName',
    '_getSession',
    '_cacheSession',
    // https://github.com/nodejs/node/pull/4982
    '_evictSession',
].forEach(function (method) {
    if (typeof OriginalHttpsAgent.prototype[method] === 'function') {
        HttpsAgent.prototype[method] = OriginalHttpsAgent.prototype[method];
    }
});

module.exports.HttpsAgent = HttpsAgent;
module.exports.HttpAgent = HttpAgent;