

var querystring = require('querystring');
var util = require('util');


function calculateAlphabetSize(password) {
    var ustr = '';
    for (var i = 0; i < password.length; i++) {
        if (ustr.indexOf(password.charAt(i)) == -1) {
            ustr += password.charAt(i);
        }
    }
    return ustr;
}

function GetCCType(cc_num) {
    var type = 0;

    if (cc_num.charAt(0) == '4' && (cc_num.length == 16 || cc_num.length == 13))
        type = 1;
    else if (cc_num.charAt(0) == '5' && cc_num.length == 16)
        type = 2;
    else if (cc_num.charAt(0) == '3' && (cc_num.charAt(1) == '4' || cc_num.charAt(1) == '7') && cc_num.length == 15)
        type = 3;
    else if (cc_num.charAt(0) == '6' && cc_num.charAt(1) == '0' && cc_num.charAt(2) == '1' && cc_num.charAt(3) == '1' && cc_num.length == 16)
        type = 4; 
    return type;

}

function luhnChk(luhnstr) {
    try {
        var luhn = luhnstr.replace(/[^0-9]/g, '');
        var len = luhn.length,
            mul = 0,
            prodArr = [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 2, 4, 6, 8, 1, 3, 5, 7, 9]
            ],
            sum = 0;
        
        if (len != 16 || ((luhn.length - len) >= 6)) {
            return false;
        }

        if (GetCCType(luhn) == 0) {
            return false;
        }

        if (calculateAlphabetSize(luhn).length < 5) {
            return false;
        }

        while (len--) {
            sum += prodArr[mul][parseInt(luhn.charAt(len), 10)];
            mul ^= 1;
        }
        return sum % 10 === 0 && sum > 0;
    } catch (exception) {
        console.error(exception)
        return false;
    }
}

function IsJsonString(str) {
    try {
        return JSON.parse(str);
    } catch (e) {
        return false;
    }
}


function checkJSONForLuhn10(jsonObject) {
    for (var i in jsonObject) {
        if (typeof (jsonObject[i]) === 'string') {
            if (luhnChk(jsonObject[i]) === true) {
                return true;
            }
        } else if (typeof (jsonObject[i]) === 'object') {
            if (checkJSONForLuhn10(jsonObject[i]) === true) {
                return true;
            }
        }
    }

    return false;
}


function checkParsedRequestStringForLuhn10(parsedBody) {
    if (!util.isUndefined(parsedBody)) {
        for (var pair in parsedBody) {
            if (luhnChk(parsedBody[pair].toString()) === true) {
                return true;
            }
        }
    }
    return false;
}



function checkLuhn10(str) {

    if (checkParsedRequestStringForLuhn10(querystring.parse(str))) {
        return true;
    }

    var possibleJson = IsJsonString(str);
    if (possibleJson) {
        if (checkJSONForLuhn10(possibleJson)) {
            return true;
        }
    }
    return false;
}

exports.checkLuhn10 = checkLuhn10;